import numpy as np
import gurobipy as gb
import time
import pickle
import pandas as pd
from xlsxwriter.workbook import Workbook

class hub_spoke():

    def __init__(self, vehicles):
        self.vehicles = vehicles
        self.c_vehicle = 1

    def readData(self, filename):
        df = pd.read_excel(filename, index_col = 0)
        self.travel_time = df.to_numpy()
        self.customers = self.travel_time.shape[0] - 1 #-1 stand for depot
        self.positions = int(np.ceil(self.customers / self.vehicles)) + 1
        self.stopName = df.columns

        return self

    def u_to_orderL(self, u):
        new_u = [u[:, :, j] for j in range(self.vehicles) if np.max(u[:, :, j]) > 0]
        vehicles = len(new_u)
        orderL = [[] for _ in range(vehicles)]
        positions = new_u[0].shape[1]
        for j in range(vehicles):
            for k in range(positions):
                orderL[j].append(int(np.where(abs(new_u[j][:,k] - 1) < 1e-3)[0]))
            if orderL[j][-1] != self.customers:
                orderL[j].append(self.customers) 
        
        return orderL


    def solveTSModel(self):

        m = gb.Model()

        #first-stage variables
        x = m.addVars(self.vehicles, lb = 0, vtype = gb.GRB.BINARY, name = "x")  #if we operate this route
        y = m.addVars(self.customers, self.vehicles, lb=0, vtype=gb.GRB.BINARY, name = "y") # if we assign  this customer to the vehicle 
        u = m.addVars(self.customers+1, self.positions, self.vehicles, lb=0, vtype=gb.GRB.BINARY, name="u") #if we assign this customer to a position on that vehicle
        z = m.addVars(self.customers+1, self.customers+1, self.positions-1, self.vehicles, vtype=gb.GRB.CONTINUOUS, name="z") 

        m.setObjective(
                    self.c_vehicle * gb.quicksum(x[j] for j in range(self.vehicles)) + 
                    gb.quicksum(gb.quicksum(self.travel_time[self.customers,i] * u[i,0,j] for i in range(self.customers)) + 
                    gb.quicksum(gb.quicksum(self.travel_time[i1,i2] * z[i1,i2,k,j] for i1 in range(self.customers+1) for i2 in range(self.customers+1)) for k in range(self.positions-1)) +
                    gb.quicksum(self.travel_time[i,self.customers]*u[i,self.positions-1,j] for i in range(self.customers)) for j in range(self.vehicles))) 
                    

        for i in range(self.customers):
            m.addConstr(
                gb.quicksum(y[i,j] for j in range(self.vehicles)) == 1)
            for j in range(self.vehicles):
                m.addConstr(
                    y[i,j] <= x[j])
                m.addConstr(
                    gb.quicksum(u[i,k,j] for k in range(self.positions)) == y[i,j])
                
        for k in range(self.positions):
            for j in range(self.vehicles):
                m.addConstr(
                    gb.quicksum(u[i,k,j] for i in range(self.customers)) + u[self.customers,k,j] == x[j])
    
                if k < self.positions-1:
                    m.addConstr(gb.quicksum(u[i,k+1,j] for i in range(self.customers)) <=
                        gb.quicksum(u[i,k,j] for i in range(self.customers)))
                    

        #constraints for McCormick envelopes
        for i1 in range(self.customers+1):
            for i2 in range(self.customers+1):
                for k in range(self.positions-1):
                    for j in range(self.vehicles):
                        m.addConstr(z[i1,i2,k,j] <= u[i1,k,j])
                        m.addConstr(z[i1,i2,k,j] <= u[i2,k + 1,j])
                        m.addConstr(z[i1,i2,k,j] >= u[i1,k,j] + u[i2,k + 1,j] - 1)

        

        m.update()

        m.setParam('TimeLimit', 2*3600)
        st = time.time()
        m.optimize()
        ed = time.time()
        print('time:', ed-st)
        opt = m.objval

        stats = {}
        stats['m_opt_u'] = np.zeros([self.customers + 1, self.positions, self.vehicles])
        stats['m_opt_x'] = np.zeros(self.vehicles)
        stats['m_opt_y'] = np.zeros([self.customers, self.vehicles])

        #stats['total_delay'] = total_delay
        stats['total_vehicle'] = self.vehicles
        stats['opt'] = opt
        #stats['MIL_gap'] = m.MIPGap
        stats['time'] = ed-st


        for i in range(self.customers + 1):
            for k in range(self.positions):
                for j in range(self.vehicles):
                    stats['m_opt_u'][i,k,j] = m.getVarByName('u[%d,%d,%d]' % (i,k,j)).x
        
        for i in range(self.customers):
            for j in range(self.vehicles):
                stats['m_opt_x'][j] = m.getVarByName('x[%d]' % j).x
                stats['m_opt_y'][i,j] = m.getVarByName('y[%d,%d]' % (i,j)).x
        
        orderL = self.u_to_orderL(stats['m_opt_u'])
        
        output = Workbook('Output.xlsx')
        output_result = output.add_worksheet('result')
        output_result.write(0, 0, 'There are ' + str(len(orderL)) + ' routes generated in total:')
        for i in range(len(orderL)):
            output_result.write(i + 1, 0, 'Route ' + str(i + 1) + ' : ' + '-> '.join([str(self.stopName[orderL[i][j]]) for j in range(len(orderL[i]))]) )
        output.close()
        
        
        return output

    


if __name__ == "__main__":
    vehicles = 3
    HubSpokeModel = hub_spoke(vehicles)
    HubSpokeModel.readData("distance.xlsx")
    results = HubSpokeModel.solveTSModel()
    print("Optimization done! The results are saved in 'Output.xlsx' file.")
